/*
  PushIt - A simple Sokoban solving game
  Copyright (C) 2006  Tim Teulings

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "GameArea.h"

#include <Lum/Base/Path.h>
#include <Lum/Base/Util.h>

#include <Lum/OS/Driver.h>

GameArea::GameArea()
 : themeIndex(0),
   xOff(0),
   yOff(0),
   finished(new Lum::Model::Boolean),
   gameBackground(NULL),
   game(NULL),
   sokobanColor("red",Lum::OS::display->GetColor(Lum::OS::Display::blackColor)),
   frameShineColor("grey95",Lum::OS::display->GetColor(Lum::OS::Display::blackColor)),
   frameShadowColor("grey20",Lum::OS::display->GetColor(Lum::OS::Display::blackColor))
{
  SetCanFocus(true);
  RequestFocus();

  finished->Set(false);
}

GameArea::~GameArea()
{
  delete gameBackground;
  delete game;
}

bool GameArea::HandlesClickFocus() const
{
  return CanFocus() && IsVisible();
}

bool GameArea::HandlesKeyFocus() const
{
  return CanFocus() && IsVisible();
}

bool GameArea::RequiresKnobs() const
{
  return false;
}

void GameArea::UpdateDimensions()
{
  if (game!=NULL) {
    hAdjustment->SetDimension(width,gameBackground->GetWidth());
    vAdjustment->SetDimension(height,gameBackground->GetHeight());
  }
  else {
    hAdjustment->SetInvalid();
    vAdjustment->SetInvalid();
  }
}

size_t GameArea::GetBoxSize() const
{
  return themes[themeIndex].boxSize;
}

Lum::Images::Image* GameArea::GetImage(Image image) const
{
  return themes[themeIndex].images[image];
}

size_t GameArea::GetSokobanXPos() const
{
  size_t sx;
  size_t sy;

  sokoban.GetSokoban(sx,sy);

  return sx;
}

size_t GameArea::GetSokobanYPos() const
{
  size_t sx;
  size_t sy;

  sokoban.GetSokoban(sx,sy);

  return sy;
}

void GameArea::MakeSokobanVisible()
{
  size_t sx;
  size_t sy;

  sokoban.GetSokoban(sx,sy);

  hAdjustment->MakeVisible(sx*GetBoxSize()+1);
  vAdjustment->MakeVisible(sy*GetBoxSize()+1);

  hAdjustment->MakeVisible((sx+1)*GetBoxSize()+1);
  vAdjustment->MakeVisible((sy+1)*GetBoxSize()+1);
}

void GameArea::MoveLeft()
{
  sokoban.MoveLeft();
  RebuildGame();
  MakeSokobanVisible();
  Redraw();

  finished->Set(sokoban.IsFinished());
}

void GameArea::MoveUp()
{
  sokoban.MoveUp();
  RebuildGame();
  MakeSokobanVisible();
  Redraw();

  finished->Set(sokoban.IsFinished());
}

void GameArea::MoveRight()
{
  sokoban.MoveRight();
  RebuildGame();
  MakeSokobanVisible();
  Redraw();

  finished->Set(sokoban.IsFinished());
}

void GameArea::MoveDown()
{
  sokoban.MoveDown();
  RebuildGame();
  MakeSokobanVisible();
  Redraw();

  finished->Set(sokoban.IsFinished());
}

void GameArea::DrawSokobanInternal(Lum::OS::DrawInfo *draw, size_t x, size_t y, size_t size)
{
  draw->PushForeground(sokobanColor);
  draw->FillRectangle(x,y,size,size);
  draw->PopForeground();

  // Left
  draw->PushForeground(frameShineColor);
  draw->DrawLine(x,y,x,y+size-1);
  /// Top
  draw->DrawLine(x,y,x+size-1,y);
  draw->PopForeground();

  // Right
  draw->PushForeground(frameShadowColor);
  draw->DrawLine(x+size-1,y,x+size-1,y+size-1);
  // Bottom
  draw->DrawLine(x,y+size-1,x+size-1,y+size-1);
  draw->PopForeground();
}

void GameArea::DrawPackageInternal(Lum::OS::DrawInfo *draw, size_t x, size_t y, size_t size)
{
  draw->PushForeground(Lum::OS::display->fillColor);
  draw->FillRectangle(x,y,size,size);
  draw->PopForeground();

  // Left
  draw->PushForeground(frameShineColor);
  draw->DrawLine(x,y,x,y+size-1);
  /// Top
  draw->DrawLine(x,y,x+size-1,y);
  draw->PopForeground();

  // Right
  draw->PushForeground(frameShadowColor);
  draw->DrawLine(x+size-1,y,x+size-1,y+size-1);
  // Bottom
  draw->DrawLine(x,y+size-1,x+size-1,y+size-1);
  draw->PopForeground();
}

void GameArea::DrawGoalInternal(Lum::OS::DrawInfo *draw, size_t x, size_t y, size_t size)
{
  draw->PushForeground(frameShadowColor);
  draw->FillRectangle(x,y,size,size);
  draw->PopForeground();

}

void GameArea::DrawFloorInternal(Lum::OS::DrawInfo *draw, size_t x, size_t y, size_t size)
{
  draw->PushForeground(Lum::OS::Display::whiteColor);
  draw->FillRectangle(x,y,size,size);
  draw->PopForeground();
}

void GameArea::DrawWallInternal(Lum::OS::DrawInfo *draw,
                                size_t x, size_t y, size_t size,
                                size_t cx, size_t cy)
{
  Lum::OS::FillRef fill=Lum::OS::display->GetFill(Lum::OS::Display::backgroundFillIndex);

  fill->Draw(draw,x,y,size,size,x,y,size,size);

  // Left
  if (cx==0 || !sokoban.IsWall(cx-1,cy)) {
    draw->PushForeground(frameShineColor);
    draw->DrawLine(x,y,x,y+size-1);
    draw->PopForeground();
  }
  /// Top
  if (cy==0 || !sokoban.IsWall(cx,cy-1)) {
    draw->PushForeground(frameShineColor);
    draw->DrawLine(x,y,x+size-1,y);
    draw->PopForeground();
  }
  // Right
  if (cx==sokoban.GetWidth()-1 || !sokoban.IsWall(cx+1,cy)) {
    draw->PushForeground(frameShadowColor);
    draw->DrawLine(x+size-1,y,x+size-1,y+size-1);
    draw->PopForeground();
  }
  // Bottom
  if (cy==sokoban.GetHeight()-1 || !sokoban.IsWall(cx,cy+1)) {
    draw->PushForeground(frameShadowColor);
    draw->DrawLine(x,y+size-1,x+size-1,y+size-1);
    draw->PopForeground();
  }
}

/**
  Create bitmap containing the game area background build up from walls, floors,
  goals and backgrounds.
 */
void GameArea::RebuildGameBackground()
{
  delete gameBackground;
  delete game;

  gameBackground=Lum::OS::driver->CreateBitmap(sokoban.GetWidth()*GetBoxSize(),
                                               sokoban.GetHeight()*GetBoxSize());

  Lum::OS::FillRef fill=Lum::OS::display->GetFill(Lum::OS::Display::backgroundFillIndex);

  fill->Draw(gameBackground->GetDrawInfo(),
             0,0,sokoban.GetWidth()*GetBoxSize(),sokoban.GetWidth()*GetBoxSize(),
             0,0,sokoban.GetWidth()*GetBoxSize(),sokoban.GetWidth()*GetBoxSize());

  game=Lum::OS::driver->CreateBitmap(sokoban.GetWidth()*GetBoxSize(),
                                     sokoban.GetHeight()*GetBoxSize());

  for (size_t cx=0; cx<sokoban.GetWidth(); ++cx) {
    for (size_t cy=0; cy<sokoban.GetHeight(); ++cy) {
      unsigned long  type=sokoban.Get(cx,cy);

      if (type & Sokoban::typeWall) {
        if (GetImage(wallImage)!=NULL) {
          GetImage(wallImage)->Draw(gameBackground->GetDrawInfo(),
                                    cx*GetBoxSize(),
                                    cy*GetBoxSize());
        }
        else {
          DrawWallInternal(gameBackground->GetDrawInfo(),cx*GetBoxSize(),cy*GetBoxSize(),GetBoxSize(),cx,cy);
        }
      }
      else if (type & Sokoban::typeGoal) {
        if (GetImage(goalImage)!=NULL) {
          GetImage(goalImage)->Draw(gameBackground->GetDrawInfo(),
                                    cx*GetBoxSize(),
                                    cy*GetBoxSize());
        }
        else {
          DrawGoalInternal(gameBackground->GetDrawInfo(),cx*GetBoxSize(),cy*GetBoxSize(),GetBoxSize());
        }
      }
      else if (type & Sokoban::typeFloor) {
        if (GetImage(floorImage)!=NULL) {
          GetImage(floorImage)->Draw(gameBackground->GetDrawInfo(),
                                     cx*GetBoxSize(),
                                     cy*GetBoxSize());
        }
        else {
          DrawFloorInternal(gameBackground->GetDrawInfo(),cx*GetBoxSize(),cy*GetBoxSize(),GetBoxSize());
        }
      }
    }
  }
}

void GameArea::RebuildGame()
{
  game->GetDrawInfo()->CopyFromBitmap(gameBackground,
                                      0,0,sokoban.GetWidth()*GetBoxSize(),sokoban.GetHeight()*GetBoxSize(),
                                      0,0);

  for (size_t cx=0; cx<sokoban.GetWidth(); ++cx) {
    for (size_t cy=0; cy<sokoban.GetHeight(); ++cy) {
      unsigned long  type=sokoban.Get(cx,cy);

      //
      // Draw foreground
      //

      if (type & Sokoban::typePackage) {
        if (GetImage(packageImage)!=NULL) {
          GetImage(packageImage)->Draw(game->GetDrawInfo(),
                                       cx*GetBoxSize(),
                                       cy*GetBoxSize());
        }
        else {
          DrawPackageInternal(game->GetDrawInfo(),cx*GetBoxSize(),cy*GetBoxSize(),GetBoxSize());
        }
      }

      if (type & Sokoban::typeSokoban) {
        if (GetImage(sokobanImage)!=NULL) {
          GetImage(sokobanImage)->Draw(game->GetDrawInfo(),
                                       cx*GetBoxSize(),
                                       cy*GetBoxSize());
        }
        else {
          DrawSokobanInternal(game->GetDrawInfo(),cx*GetBoxSize(),cy*GetBoxSize(),GetBoxSize());
        }
      }
    }
  }
}

void GameArea::EvaluateThemeName()
{
  for (size_t i=0; i<themes.size(); i++) {
    if (themes[i].name==themeName) {
      themeIndex=i;

      xOff=0;
      yOff=0;

      // Only rebuild game area if there is a valid Sokoban assigned!
      if (sokoban.GetWidth()>0) {
        RebuildGameBackground();
        RebuildGame();
        Redraw();
      }
      break;
    }
  }
}

void GameArea::SetSokoban(const Sokoban& sokoban)
{
  this->sokoban=sokoban;

  xOff=0;
  yOff=0;

  RebuildGameBackground();
  RebuildGame();
  UpdateDimensions();
  Redraw();
  finished->Set(false);
}

void GameArea::CalcSize()
{
  if (Lum::OS::display->GetSize()<Lum::OS::Display::sizeNormal) {
    width=10*32;
    height=10*32;
  }
  else {
    width=20*32;
    height=20*32;
  }

  minWidth=10*32;
  minHeight=10*32;

  Scrollable::CalcSize();
}

void GameArea::GetDimension(size_t& width, size_t& height)
{
  if (gameBackground!=NULL) {
    width=gameBackground->GetWidth();
    height=gameBackground->GetHeight();
  }
}

void GameArea::Layout()
{
  UpdateDimensions();

  Scrollable::Layout();
}

void GameArea::Draw(Lum::OS::DrawInfo* draw,
                    int x, int y, size_t w, size_t h)
{
  Scrollable::Draw(draw,x,y,w,h);

  if (!OIntersect(x,y,w,h)) {
    return;
  }

  /* --- */

  if (game==NULL) {
    return;
  }

  draw->CopyFromBitmap(game,
                       hAdjustment->GetTop()-1,vAdjustment->GetTop()-1,
                       width,height,
                       this->x,this->y);
}

bool GameArea::HandleMouseEvent(const Lum::OS::MouseEvent& event)
{
  if (event.type==Lum::OS::MouseEvent::down && PointIsIn(event) && event.button==Lum::OS::MouseEvent::button1) {
    size_t x,y;
    size_t sx,sy;
    size_t moves;

    x=(event.x-this->x+hAdjustment->GetTop()-1)/GetBoxSize();
    y=(event.y-this->y+vAdjustment->GetTop()-1)/GetBoxSize();

    sokoban.GetSokoban(sx,sy);

    if ((sx==x && sy==y)/*  || (sx!=x && sy!=y)*/) {
      // No move requested
      return true;
    }

    if (sx==x) {
      // Vertical movement
      if (y>sy) {
        while (y>GetSokobanYPos() && sokoban.CanMoveDown()) {
          MoveDown();
        }
      }
      else {
        while (y<GetSokobanYPos() && sokoban.CanMoveUp()) {
          MoveUp();
        }
      }
    }
    else if (sy==y) {
      // Horizontal movement
      if (x>sx) {
        while (x>GetSokobanXPos() && sokoban.CanMoveRight()) {
          MoveRight();
        }
      }
      else {
        while (x<GetSokobanXPos() && sokoban.CanMoveLeft()) {
          MoveLeft();
        }
      }
    }

    sokoban.GetSokoban(sx,sy);

    if ((sx!=x || sy!=y) && sokoban.IsConnected(sx,sy,x,y,moves)) {
      sokoban.MoveSokoban(x,y);
      RebuildGame();
      MakeSokobanVisible();
      Redraw();

      finished->Set(sokoban.IsFinished());
    }


    return true;
  }

  return false;
}

bool GameArea::HandleKeyEvent(const Lum::OS::KeyEvent& event)
{
  if (event.type==Lum::OS::KeyEvent::down) {
    switch (event.key) {
    case Lum::OS::keyLeft:
      if (!sokoban.CanMoveLeft()) {
        Lum::OS::display->Beep();
      }
      else {
        MoveLeft();
        finished->Set(sokoban.IsFinished());
      }
      return true;
    case Lum::OS::keyRight:
      if (!sokoban.CanMoveRight()) {
        Lum::OS::display->Beep();
      }
      else {
        MoveRight();
        finished->Set(sokoban.IsFinished());
      }
      return true;
    case Lum::OS::keyUp:
      if (!sokoban.CanMoveUp()) {
        Lum::OS::display->Beep();
      }
      else {
        MoveUp();
        finished->Set(sokoban.IsFinished());
      }
      return true;
    case Lum::OS::keyDown:
      if (!sokoban.CanMoveDown()) {
        Lum::OS::display->Beep();
      }
      else {
        MoveDown();
        finished->Set(sokoban.IsFinished());
      }
      return true;
    case Lum::OS::keyBackspace:
    case Lum::OS::keyDelete:
      Undo();
      return true;
    default:
      return false;
    }
  }

  return false;
}

Lum::Model::Boolean* GameArea::GetFinishedModel() const
{
  return finished.Get();
}

void GameArea::Undo()
{
  sokoban.Undo();
  RebuildGame();
  Redraw();
  finished->Set(sokoban.IsFinished());
}

void GameArea::Resync(Lum::Base::Model* model, const Lum::Base::ResyncMsg& msg)
{
  if (model==hAdjustment->GetTopModel() || model==vAdjustment->GetTopModel()) {
    Redraw();
  }

  Lum::Scrollable::Resync(model,msg);
}

